# -*- coding: utf-8 -*-
"""
Created on Thu Jun  3 15:19:21 2021

@author: Urs


Drehimpulserhaltung ist nicht umgesetzt, Impulserhaltung eigentlich schon, 
der Energieumsatz pro Reaktion ist zufällig, variabel

Energieerhaltung lässt sich ohne Stosspartner oder innere Freiheitsgrade auch gar nicht umsetzen, denn bei einem inelastischen Stoss
geht - ausser in ganz speziellen Fällen -  Energie verloren, mehr Energie wird es sicher nicht, aber hier ist die Bindungsbildung 
zwischen den beiden NO2 eine exotherme Reaktion. Diese Energie kann also nicht direkt in Form von translationsbewegung frei werden, 
sondern nur über Schwingungen, Drehungen, angeregte Elektronen oder Stosspartner. 




ffmpeg -start_number 0 -i pygame_%04d.png -filter:v fps=30  -vf "pad=ceil(iw/2)*2:ceil(ih/2)*2" -pix_fmt yuv420p pytame_out.mp4
ffmpeg -start_number 0 -i matplotlib_%04d.png -filter:v fps=30  -vf "pad=ceil(iw/2)*2:ceil(ih/2)*2" -pix_fmt yuv420p matplotlib_out.mp4



"""

# ---------------------------------------------------------------------------
# Externe Bibliotheken einbinden
import pygame as pg
import random as rd
import sys
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
import os

# ---------------------------------------------------------------------------
# Wichtige Variabeln

#Grösse der dargestellten Bilder
molekuelbreite = 15                                            # Breite (pixel) der Molekplüle (10 oder 30)

# Dimension des pygame-Fensters
fenster_b = 800
fenster_h = fenster_b    
# Dimensionen der Teilchenbegrenzung (box)
boxdelta = fenster_b/2-80

boxzentrum = fenster_b/2
boxdelta_max = fenster_b/2-10

# Dimensionen des Matplotlib-Fensters
figsize = (8,8)
q_darstellen = True                                           # mit False werden nur die Konzentrationen dargestellt 

# Anzahl Edukte und Produkte
n_no2 = 0
n_n2o4 = 40                                                 # z.B. 10 oder 40

 
# Geschwindigkeitskonstante der Hin- und Rückreaktion eigentlich ist das 
# die Wahrscheinlichkeit einer Reaktion in einem Schritt pro Teilchen (rück) oder pro Zusammenstoss (hin)
k_hin = 0.01
k_rueck = 0.005

v_faktor  = 8                                                   # Anfangsgeschwindigkeit (z.B. 8 oder 20)

tastenbefehle_anzeigen = True

bilder_speichern = False

# Farben
SCHWARZ     = [0,0,0]
WEISS       = [255,255,255]
ROT         = [255,0,0]
BLASSROT    = [255,180,180]
GRUEN       = [0,255,0]
BLAU        = [0,0,255]
GELB        = [255,255,0]
GRAU        = [100, 100, 100]
NAVYBLUE    = [60, 60, 100]
ORANGE      = [255, 128, 0]
MAGENTA     = [255, 0, 255]
CYAN        = [0, 255, 255]




# ----------------------------------
# Matplotlib-Fenster vorbereiten
def move_figure(f, x, y):
    """Move figure's upper left corner to pixel (x, y)"""
    backend = matplotlib.get_backend()
    if backend == 'TkAgg':
        f.canvas.manager.window.wm_geometry("+%d+%d" % (x, y))
    elif backend == 'WXAgg':
        f.canvas.manager.window.SetPosition((x, y))
    else:
        # This works for QT and GTK
        # You can also use window.setGeometry
        f.canvas.manager.window.move(x, y)

plt.ion()
if q_darstellen == False:
    fig, ax = plt.subplots(1,1, figsize = figsize)
    ax0 = ax
    ax0.set_xlim([0,510])
    ax0.set_xticks([])
else:
    fig, ax = plt.subplots(2,1, figsize = figsize)
    ax0 = ax[0]
    for a in ax:
        a.set_xlim([0,510])
        a.set_xticks([]) 
    linie_verh, = ax[1].plot([],[], 'k--')
    linie_q, = ax[1].plot([],[], 'k-')
    ax[1].set_ylim([0,5])
    ax[1].set_xlabel("Zeit")
    ax[1].legend([r"$\frac{[N_2O_4]}{[NO_2]}$", r"$Q = \frac{[N_2O_4]}{[NO_2]^2}$"])

x_window = 40 + fenster_b
y_window = 20
move_figure(fig, x_window, y_window)

m_NO2 = 32+14
m_N2O4 = 2*(32+14)

# ---------------------------------------------------------------------------
# Pygame initialisieren
x_window = 20
y_window = 50
os.environ['SDL_VIDEO_WINDOW_POS'] = "%d,%d" % (x_window,y_window)
pg.init()                                           # Pygame initialisieren
screen = pg.display.set_mode((fenster_b,fenster_h)) # Fenstergrösse setzen
pg.display.set_caption("Gleichgewichtsreaktion")    # Titel des Fensters
my_font = pg.font.SysFont('Comic Sans MS', 10)      # Font zum schreiben laden




# ---------------------------------------------------------------------------
# Klassendefinitionen

class Box:
    def __init__(self,delta = 330, zentrum = 350, delta_max = 340):
        self.delta_max = delta_max
        self.delta = delta
        self.zentrum = zentrum
        self.resize()
    def resize(self):
        self.range=np.array([self.zentrum-self.delta, self.zentrum+self.delta, self.zentrum-self.delta, self.zentrum+self.delta])   #xmin, xmax, ymin, ymax
    def update(self):
        pg.draw.rect(screen, WEISS, (self.range[0],self.range[2], 2*self.delta, 2*self.delta), 3)

class Molekuel(pg.sprite.Sprite):
    def __init__(self, typ=0, pos = [10,10], v = [1,1], omega = 1, ident = 0, m=1):
        super().__init__()
        self.typ = typ
        self.ident  = ident 
        if typ == 0:
            self.bild = bilder.bild[0]
            self.m = m_NO2
        elif typ == 1:
            self.bild = bilder.bild[1]
            self.m = m_N2O4
        self.image = self.bild 
        self.omega = omega                                    #winkelgeschwindigkeit
        self.pos = np.array(pos)
        self.v = np.asarray(v)
        self.rect = self.image.get_rect()
        self.rect.center = self.pos
        self.gravity = np.array([0,0])
        self.v = np.asarray(v)
        #self.music = pg.mixer.Sound("res/whoosh.wav")
        self.t_drehung = 0
          
    def update(self):
        self.v = self.v + self.gravity
        self.v_betrag = np.linalg.norm(self.v)
        if self.v_betrag > 20:
            self.v = self.v/self.v_betrag * 20
        self.rect.center += self.v
        
        if self.rect[0] < box.range[0]:
            self.rect[0] = box.range[0] + self.rect.width *0.01 
            self.v[0] = self.v[0]  * -1 
        if self.rect[0] > box.range[1] - self.rect.width:
            self.rect[0] = box.range[1] - self.rect.width * 1.01 
            self.v[0] = self.v[0]  * -1                               
        if self.rect[1] < box.range[2]: 
            self.rect[1] = box.range[2] + self.rect.height * 0.01
            self.v[1] = self.v[1]  * -1       
        if self.rect[1] > box.range[3] - self.rect.height:
            self.rect[1] = box.range[3]  - self.rect.height*1.01
            self.v[1] = self.v[1]  * -1
        
        # Drehung des Bildes
        # Da ein Bild nicht mehrfach gedreht werden kann - es wird dabei völlig zerrissen
        # muss dasselbe ursprüngliche Bild immer wieder gedreht werden
        self.t_drehung += 1
        self.pos = np.array(self.rect.center)
        self.bildTransf = pg.transform.rotate(self.bild, self.omega*self.t_drehung)
        self.rect1 = self.bildTransf.get_rect()
        self.image = self.bildTransf
        self.rect = self.image.get_rect()
        self.rect.center = self.pos



        
class Bilder():
     def __init__(self, molekuelbreite  = 20):
        # mit png wird der (eigentlich transparente) Hintergrund schwarz
        bild_dateien = ["resources/no2.png", "resources/n2o4.png"]        
        self.bild = []
        new_width = molekuelbreite 
        for b in bild_dateien:
            bild = pg.image.load(b).convert_alpha()
            
            new_height = int(bild.get_height()/bild.get_width() * new_width)
            bild = pg.transform.smoothscale(bild, (new_width, new_height))
            self.bild.append(bild)
            

class einzelKollision():
    def __init__(self,p1,p2,v1,v2, dmin=5, m1=1, m2=1, dt = 1):
        """
        Die beiden Methoden - Rechtecks-Kollision und Reflexion - beissen sich 
        (z.B. wenn ein Teilchen nach der einen Methode kollidiert und nach der anderen nicht), 
        so dass die Teilchen manchmal ruckeln.

        """
        self.dt = dt
        self.p1 = np.array(p1,dtype=float)
        self.p2 = np.array(p2,dtype=float)
        self.v1 = np.array(v1,dtype=float)
        self.v2 = np.array(v2,dtype=float)
        d = self.p2-self.p1

        if np.linalg.norm(d) <= dmin and np.linalg.norm(d) != 0: #sonst gibt es bei der Kollsion Fehler
            # location vector
            r1 = self.p1
            r2 = self.p2

            # velocity vector
            v1 = self.v1
            v2 = self.v2

            # relative location & velocity vectors
            r_rel = r1 - r2
            v_rel = v1 - v2

            # momentum vector of the center of mass
            v_cm = (m1 * v1 + m2 * v2) / (m1 + m2)

            # collisions of spheres reflect v_rel over r_rel
            r_rel_ev = r_rel / (np.dot(r_rel, r_rel))**0.5           # Einheitsvektor in Richtung des Verbindungsvektors
            vr_rel = np.dot(v_rel, r_rel_ev)
            v_rel = 2 * r_rel_ev * vr_rel - v_rel

            # assign new velocities #falsch herum im Original
            self.v1 = v_cm - v_rel * m2 / (m1 + m2)
            self.v2 = v_cm + v_rel * m1 / (m1 + m2) 

            #Teilchen auseinanderrücken: entlang der neuen Geschwindigkeiten einen zeitschritt ausführen
            self.p1 += dt * self.v1
            self.p2 += dt * self.v2

            #falls Teilchen (von Beginn weg) ineinander verkeilt sind:
            d_paar  = np.linalg.norm(self.p1 - self.p2) 
            if d_paar < dmin *0.8 :
                r_cm = (r1 + r2)/2
                self.p1 = r_cm + (dmin - d_paar)/2 * r_rel_ev * 3
                self.p2 = r_cm - (dmin - d_paar)/2 * r_rel_ev * 3

        #positionen und Geschwindigkeiten gespeichert in  self.p1, self.p2, self.v1, self.v2


class einzelKollisionEnergieAenderung():
    def __init__(self,p1,p2,v1,v2, dmin=5, m1=1, m2=1, m1_neu = 1, m2_neu = 1, dt = 1, dEA = 100, dER = -100):
        """
        Die beiden Methoden - Rechtecks-Kollision und Reflexion - beissen sich 
        (z.B. wenn ein Teilchen nach der einen Methode kollidiert und nach der anderen nicht), 
        so dass die Teilchen manchmal ruckeln.

        """
        #print("massen kollision und umwandlung",m1,m2)
        self.dt = dt
        self.p1 = np.array(p1,dtype=float)
        self.p2 = np.array(p2,dtype=float)
        self.v1 = np.array(v1,dtype=float)
        self.v2 = np.array(v2,dtype=float)
        d = self.p2-self.p1
        self.umwandlung = False

        if np.linalg.norm(d) <= dmin and np.linalg.norm(d) != 0: #sonst gibt es bei der Kollsion Fehler
            # location vector
            r1 = self.p1
            r2 = self.p2

            # velocity vector
            v1 = np.copy(self.v1)
            v2 = np.copy(self.v2)

            # relative location & velocity vectors
            r_rel = r1 - r2
            v_rel = v1 - v2

            # momentum vector of the center of mass
            v_cm = (m1 * v1 + m2 * v2) / (m1 + m2)
            
            
            # collisions of spheres reflect v_rel over r_rel
            r_rel_ev = r_rel / (np.dot(r_rel, r_rel))**0.5          # Einheitsvektor in Richtung des Verbindungsvektors
            vr_rel = np.dot(v_rel, r_rel_ev)                        # Relativgeschwindigkeit projiziert auf den Verbindungsvektor 
            v_rel = 2 * r_rel_ev * vr_rel - v_rel                   # 
            v1_rel = - v_rel * m2 / (m1 + m2)
            v2_rel = + v_rel * m1 / (m1 + m2) 
            
            # Ekin der Stosspartner im Inertialsystem des Schwerpunkts
            """
            Hier wird die Richtungsänderung unter Erhaltung des Impulses berechnet
            ENERGIEERHALTUNG IST ABER VERLETZT!!
            Wird anschliessend korrigiert, indem die Geschwindigkeit wieder so verteilt wird, dass auch die Energie stimmt
            Vielleicht GEHT das gar nicht mit Energieerhaltung!!!!!!!!!!!
            Denn schliesslich werden Atome umgehängt und kleben anschliessend anders aneinander, wie bei einem inelastischen Stoss!!!!!
            """
            
            
            Ekin_tot_in = 0.5 *(m1 * (np.linalg.norm(v1_rel))**2 + m2 * (np.linalg.norm(v2_rel))**2)
            if Ekin_tot_in > dEA: 
                v1 = v1 * m1 /m1_neu   #Teilchen werden für einen korrekten Stoss schon vor der Kollision verwandelt mit Impulserhaltung
                v2 = v2 * m2 /m2_neu 
                m1 = m1_neu
                m2 = m2_neu
                v_rel = v1 - v2
                vr_rel = np.dot(v_rel, r_rel_ev)                        # Relativgeschwindigkeit projiziert auf den Verbindungsvektor 
                v_rel = 2 * r_rel_ev * vr_rel - v_rel                   # 
                v1_rel = - v_rel * m2 / (m1 + m2)
                v2_rel = + v_rel * m1 / (m1 + m2)                  


            # assign new velocities #falsch herum im Original
            v1_rel = - v_rel * m2 / (m1 + m2)
            v2_rel = + v_rel * m1 / (m1 + m2) 
            self.v1 = v_cm + v1_rel
            self.v2 = v_cm + v2_rel



            # Ekin der stosspartner im ruhenden System
            Ekin_tot_1 = 0.5 *(m1 * (np.linalg.norm(v1))**2 + m2 * (np.linalg.norm(v2))**2)
            Ekin_tot_2 = np.sum(0.5 *(m1 * self.v1**2 + m2 * self.v2**2))
            
            # print("ekin mittel mal 2", e_kin_mittel()*2)
            # print("Umwandlung: summe der Energien der Stosspartner")
            # print("Ekin nach reflexion vor Reaktion", Ekin_tot_1, Ekin_tot_2)
            
            if Ekin_tot_in > dEA:   #wenn die Aktivierungsenergie überschritten ist
                #print("Änderung, Ekin_tot, dEA, dER", Ekin_tot, dEA, dER)
                print()
                print("Umwandlung geschieht mit deltaEa, deltaEr", dEA, dER)
                print("Ekin inertial vorher ", Ekin_tot_in)
                Eneu = Ekin_tot_in - dER
                print("Eneu              ", Eneu)
                print("v1, v2 alt", self.v1, self.v2)
                v1 = (2 * Eneu * m2/ m1 / (m1 + m2))**0.5
                v2 = (2 * Eneu * m1 / m2/ (m1 + m2))**0.5
                self.v1 = v_cm + v1 * v1_rel / np.linalg.norm(v1_rel)
                self.v2 = v_cm + v2 * v2_rel / np.linalg.norm(v2_rel)
                self.umwandlung = True
               
                print("v1, v2 neu", self.v1, self.v2)
                Ekin_tot_in = 0.5 *(m1 * (np.linalg.norm(self.v1-v_cm))**2 + m2 * (np.linalg.norm(self.v2-v_cm))**2)
                print("Ekin inertial neu", Ekin_tot_in)
                Ekin_tot= 0.5 *(m1 * np.linalg.norm(self.v1)**2 + m2 * np.linalg.norm(self.v2)**2)
                print("Ekin neu                      ", Ekin_tot)
                print()
        #positionen und Geschwindigkeiten gespeichert in  self.p1, self.p2, self.v1, self.v2


            #Teilchen auseinanderrücken: entlang der neuen Geschwindigkeiten einen zeitschritt ausführen
            self.p1 += dt * self.v1
            self.p2 += dt * self.v2

            #falls Teilchen (von Beginn weg) ineinander verkeilt sind:
            d_paar  = np.linalg.norm(self.p1 - self.p2) 
            if d_paar < dmin *0.9 :
                print("ruecken!!!!!!!!!!!!!!")
                r_cm = (r1 + r2)/2
                self.p1 = r_cm + (dmin - d_paar)/2 * r_rel_ev * 2
                self.p2 = r_cm - (dmin - d_paar)/2 * r_rel_ev * 2

# ---------------------------------------------------------------------------
# Funktionen

# Events abhandeln

pause = False
def key_handler():
    global bilder_speichern, pause, zeit, n_no2_plot, n_n2o4_plot, V_plot
    key = pg.key.get_pressed()                                           # Alle gedrückten Tasten abrufen
    
    if key[pg.K_w] == True:                                # Wurde die Taste a gedrückt?
        if box.delta < box.delta_max:
            box.delta += 1
            box.resize()
    
    if key[pg.K_s] == True:
        if box.delta > 10:                                        
            box.delta -= 1
            box.resize()

    if key[pg.K_UP] == True:                                # Wurde die Pfeiltaste nach oben gedrückt?
        try:
            ylims = ax[1].get_ylim()
            ax[1].set_ylim([0, 1.25*ylims[1]])
        except:
            print("Kein skalierbares Diagramm vorhanden")
    
    if key[pg.K_DOWN] == True:
        try:
            ylims = ax[1].get_ylim()
            ax[1].set_ylim([0, 0.8*ylims[1]])
        except:
            print("Kein skalierbares Diagramm vorhanden")

    if key[pg.K_b] == True:                                # Wurde die Taste a gedrückt?
        bilder_speichern = True
        print("bilder_speichern = ", bilder_speichern)

    if key[pg.K_n] == True:                                # Wurde die Taste a gedrückt?
        bilder_speichern = False
        print("bilder_speichern = ", bilder_speichern)    
        
    if key[pg.K_p] == True:                                # Wurde die Taste a gedrückt?
        pause = True #pause^1
        print ("Pause = ", pause)

    if key[pg.K_o] == True:                                # Wurde die Taste a gedrückt?
        pause = False#pause^1
        print ("Pause = ", pause)


    if key[pg.K_r] == True:                                # Wurde die Taste a gedrückt?
        n_durchschnittsbildung = 1
        zeit = [0]
        n_no2_plot = [n_no2_plot[-1]]
        n_n2o4_plot = [n_n2o4_plot[-1]]
        V_plot = [V_plot[-1]]
        


# diese Funktion ist nötig, umd die Geschwindigkeit der Teilchen immer etwa konstant zu halten. Bei Reaktionen verändert sich sonst die kinteitsche Energie    
def e_kin_mittel():
    teilchenliste = list(all_sprites)
    e_kin = 0
    for i in range(len(teilchenliste)):
        e_kin += np.sum(teilchenliste[i].v**2)
    e_kin_avg = e_kin/len(teilchenliste)
    print("mittlere E_kin", e_kin_avg)
    return e_kin_avg

# ---------------------------------------------------------------------------
# Globale Variabeln initialisieren

# in der Klasse bilder werden die Abbildungen für die Teilchen geladen
bilder = Bilder(molekuelbreite  = molekuelbreite )



# Die Box legt die Ränder fest, an denen die Teilchen reflektiert werden
box = Box(delta = boxdelta, zentrum = boxzentrum, delta_max = boxdelta_max)                 # zentrum ist die x- und y-Koordinate der Mitte der Box, delta die halbe Seitenlänge

game_is_running = True                                                 # Variable für den Zustand des Spiels
clock = pg.time.Clock()                                                # Eine Pygame-Uhr um die Framerate zu kontrollieren

no2_group = pg.sprite.Group()                                          # In dieser Gruppe sind alle NO2-Sprites
n2o4_group = pg.sprite.Group()                                         # In dieser Gruppe sind alle N2O4-Sprites
all_sprites = pg.sprite.Group()                                        # In dieser Gruppe sind alle Sprites die gezeichnet werden sollen


# Gesamtzahlen der Teilchen
n_teilchen = n_no2 + n_n2o4
n_teilchen_max = n_no2 + 2 * n_n2o4                     # mehr Teilchen können es unmöglich werden


# anfängliche Werte der Sprites: Positionm, Geschwindigkeit, Drehgeschwindigkeit
init_pos = (np.random.random((n_teilchen, 2))) * np.array([fenster_b*0.8, fenster_h*0.8]) + np.array([20,20])
init_v = (-0.5 + np.random.random((n_teilchen, 2)))*v_faktor 
init_omega = (-0.5 + np.random.random((n_teilchen, 1)))*6

i_m = 0

# Moleküle erzeugen und in Gruppen einspeichern
for i in range(n_no2):
    neues_molekuel = Molekuel(typ = 0, pos = init_pos[i_m], v = init_v[i_m], omega = init_omega[i_m])
    no2_group.add(neues_molekuel)  
    all_sprites.add(neues_molekuel)    
    i_m += 1          
for i in range(n_n2o4):
    neues_molekuel = Molekuel(typ = 1, pos = init_pos[i_m], v = init_v[i_m], omega = init_omega[i_m])
    n2o4_group.add(neues_molekuel)  
    all_sprites.add(neues_molekuel) 
    i_m += 1    


# ursprüngliche Ekin ermitteln. Auf diesen Wert werden die Energien der Teilchen später eingepegelt
E_kin_0 = e_kin_mittel()


# ----------------------------------
# hier wird das Diagramm vorbereitet
ax0.set_ylim([0, 1.1 * n_teilchen_max ])   
ax0.set_xlabel("Zeit")
ax0.set_ylabel("Anzahl")


linie_no2, = ax0.plot([],[])
linie_n2o4, = ax0.plot([],[])

ax0.legend([r"$n(NO_2)$", r"$n(N_2O_4)$"])

n_no2_plot = [n_no2]
n_n2o4_plot = [n_n2o4]
V_plot = [(box.delta/10)**2]
n_durchschnittsbildung = 1
n_einzelwerte = 0
n_no2_zwischen = []
n_n2o4_zwischen = []
V_zwischen = []

# ---------------------------------------------------------------------------
# Hauptschleife 

zeit = [0]
i_bild = 0
while game_is_running:                              # Solange das Spiel läuft...
    # -----------------------------------------------------------------------
    # Events und Tasten abfragen
    
                
    for event in pg.event.get():                         # Gehe alle Events des Sysems durch
        if event.type == pg.QUIT or (event.type == pg.KEYDOWN and event.key == pg.K_ESCAPE):
            game_is_running = False                      # Wenn ja, beende die Hauptschleife
                
    key_handler()     

    if pause == False:
                                   # Alle Tasten abhandeln
    
    
        # -----------------------------------------------------------------------
        # Kollisionen. Bei einer Kollision müssen die Teilchen voneinander abprallen oder miteinander reagieren (Hinreaktion)    
    
        teilchenliste = list(all_sprites)
        for i in range(len(all_sprites)-1):
            test_t = teilchenliste[i]                        # Testteilchen: getestet wird noch nicht getestete Kollision dieses Teilchens mit anderen (Testgruppe) 
            test_g = pg.sprite.Group()                       # Testgruppe
            test_g.add(teilchenliste[i+1:])
            hits = pg.sprite.spritecollide(test_t, test_g, False)
            if len(hits) > 0:
                for hit in hits:
                    if test_t.typ == 0 and hit.typ == 0 and np.random.random() < k_hin:   # Reaktion
                        zentrum = (hit.pos + test_t.pos)/2                                # Werte des alten Teilchens werden gespeichert
                        v = (hit.v + test_t.v)/2                                    
                        hit.kill()                                                        # altes Teilchen wird gelöscht
                        test_t.kill()
                        neues_molekuel = Molekuel(typ = 1, pos = zentrum, v = v, omega = (-0.5 + np.random.random())*10)
                        n2o4_group.add(neues_molekuel)  
                        all_sprites.add(neues_molekuel)                     
                    else:                                                                 # Stoss
                        koll = einzelKollision(test_t.rect.center,hit.rect.center, test_t.v, hit.v, dmin=(hit.rect.width + test_t.rect.width)/2.5, m1=1, m2=1)
                        test_t.rect.center = koll.p1
                        test_t.v = koll.v1
                        hit.rect.center = koll.p2
                        hit.v = koll.v2
        
        # Rückreaktion
        teilchenliste = list(n2o4_group)
        for i in range(len(teilchenliste)):
            if np.random.random() < k_rueck:
                v = teilchenliste[i].v
                pos = teilchenliste[i].pos
                delta_vektor = np.random.random(2)*2
                neues_molekuel = Molekuel(typ = 0, pos = pos + delta_vektor*10, v = v + delta_vektor*2, omega = (-0.5 + np.random.random())*10)
                no2_group.add(neues_molekuel) 
                all_sprites.add(neues_molekuel)               
                neues_molekuel = Molekuel(typ = 0, pos = pos - delta_vektor*10, v = v - delta_vektor*2, omega = (-0.5 + np.random.random())*10)
                no2_group.add(neues_molekuel) 
                all_sprites.add(neues_molekuel)              
                teilchenliste[i].kill()
        
        # wenn mehrere Teilchen gleichzeitig kollidieren, kann es zu einer Vermehrung der Teilchen kommen. Dies wird hier korrigiert
        while len(n2o4_group)*2 + len(no2_group) > n_teilchen_max:
            teilchenliste = list(n2o4_group)
            teilchenliste[-1].kill()
             
        # -----------------------------------------------------------------------
        # Modell zeichnen
        screen.fill(SCHWARZ)                              # Fülle den Hintergrund mit schwarzer Farbe auf
        all_sprites.update()                            # Ruft alle update Funktionen der Sprites in dieser Gruppe auf
        all_sprites.draw(screen)                        # Zeichnet automatisch alle Sprites in dieser Gruppe
        box.update()                                    # Linie der Box
        if tastenbefehle_anzeigen == True:
            label = my_font.render("Bei aktiviertem Pygame-Fenster skalieren: Box: Tasten s und w,  zweites Diagramm: up und down, speichern: b und n, Pause: p, o, Reset r", 1, WEISS)
            screen.blit(label, (5, 5))
    
        pg.display.update()                             # Zeichne den Bildschirm neu
        clock.tick(24)                                  # Der Bildschirm soll alle 1/24 Sekunden aktualisiert werden

    
        # -----------------------------------------------------------------------
        # Verlauf einzeichnen
        
           
        n_no2_zwischen.append(len(no2_group))
        n_n2o4_zwischen.append(len(n2o4_group))
        V_zwischen.append((box.delta/10)**2)
        n_einzelwerte += 1
        
        if n_einzelwerte == n_durchschnittsbildung:
            n_no2_plot.append(np.average(np.array(n_no2_zwischen)))
            n_n2o4_plot.append(np.average(np.array(n_n2o4_zwischen)))
            V_plot.append(np.average(np.array(V_zwischen)))
            linie_no2.set_data(zeit, n_no2_plot)
            linie_n2o4.set_data(zeit, n_n2o4_plot)
            n_no2_zwischen = []
            n_n2o4_zwischen = []
            V_zwischen = []
            n_einzelwerte = 0
            zeit.append(zeit[-1] +1)
    
        if q_darstellen == True:
            verhaeltnis = np.array(n_n2o4_plot)/np.array(n_no2_plot)      # Volumen kürzt sich weg
            Q = np.array(n_n2o4_plot)/np.array(n_no2_plot)**2 * V_plot    # V_plot: Einbezug des Volumens
            linie_verh.set_data(zeit, verhaeltnis)
            linie_q.set_data(zeit, Q)
    
        #wenn die Listen sehr lang werden: Einträge halbieren
        if len(zeit) == 500:
            zeit = list(np.array(zeit).reshape(2,-1)[0])
            n_no2_plot = list(np.average(np.array(n_no2_plot).reshape(-1,2), axis=1)) 
            n_n2o4_plot = list(np.average(np.array(n_n2o4_plot).reshape(-1,2), axis=1)) 
            V_plot = list(np.average(np.array(V_plot).reshape(-1,2), axis=1)) 
            n_durchschnittsbildung *= 2
            
        if len(zeit) % 20 == 0:
            #Ekin konstant halten
            if e_kin_mittel() > E_kin_0:
                for sprit in all_sprites:
                   sprit.v *= 0.995 
            else:
                for sprit in all_sprites:
                   sprit.v *= 1.005         
        plt.draw()
    
        if bilder_speichern == True:
            pg.image.save(screen, "pygame_{:04d}.png".format(i_bild))
            plt.savefig("matplotlib_{:04d}.png".format(i_bild))
            i_bild += 1
    

    
##########################################
# Hauptschleife beendet -> Game Over
##########################################

##########################################
# Spiel ist beendet
##########################################
pg.quit()                                           # Schliesse das Fenster und beende das Programm
sys.exit()
   
